/*********************************************************************
 *
 *                  BMS EEPROM File
 *
 *********************************************************************
 * FileName:        BMS_EEPROM.c
 * Processor:       PIC18F25K80
 * Compiler:        Microchip C18 v3.41
 * Company:         KIT - CN - IPE
 *
 * Author           Date         Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Ott W.           23.03.2016   Release
 *********************************************************************/

/*********************************************************************
 *
 *                          Include Files
 *
 ********************************************************************/
#include "BMS_UI_Main.h"

/*********************************************************************
*
*                             Globals
*
*********************************************************************/
extern SERIAL_NR_t  gSerial;


/*********************************************************************
 * Function:        unsigned char EEread(unsigned char adress)
 *
 * Overview:       Use this function to read EEPROM
 *
 * PreCondition:    None
 *
 * Input:           EEPROM adress
 *
 * Output:          return: EEDATA
 *
 * Side Effects:    against switch on BMS
 **********************************************************************/
uint8_t EEread(uint8_t adress)
{
    uint8_t i;
    uint8_t array_len = 0;

    array_len = sizeof(gSerial.EE_ADRESS);

    i = adress;
    for(i = 0; i < array_len;i++)
    {
        EEADR = i;                      //Address to be read
        EECON1bits.EEPGD = 0;           //Selecting EEPROM Data Memory
        EECON1bits.RD = 1;              //Initialise read cycle
        gSerial.EE_ADRESS[i] = EEDATA;
    }
    
    gSerial.EE_ADRESS[13] = 0xF0;      // Set UI Nr. to 15;
    return gSerial.EE_BYTE.FLASH;
}

/*********************************************************************
 * Function:        void EEwrite(unsigned char adress, unsigned char data)
 *
 * Overview:        Use this function to write data to EEPROM
 *
 * PreCondition:    None
 *
 * Input:           EEPROM adress, data
 *
 * Output:          None
 *
 * Side Effects:    None
 **********************************************************************/
void EEwrite(uint8_t adress, uint8_t *DATA_temp)
{

  uint8_t   INTCON_SAVE;          //To save INTCON register value
  uint8_t   i = 0;
  uint8_t   array_len = 0;
  uint16_t  LFT1_h = 0;
  uint16_t  LFT2_h = 0;
  uint8_t   BAU_h = 0;
  uint8_t   FAM_h = 0;

  if(adress == EE_308_a)
  {
      // Save the first 8 Bytes in gSerial.EE_ADRESS[1..8]
      for(i = 0; i < 8; i++)
      {
          gSerial.EE_ADRESS[i+1] = DATA_temp[i];
      }
  }
  else if(adress == EE_309_a)
  {
      // Save the second 4 Bytes in gSerial.EE_ADRESS[9..12]
      for(i = 0; i < 4; i++)
      {
          gSerial.EE_ADRESS[9+i] = DATA_temp[i];
      }

        gSerial.EE_ADRESS[0] = EE_LOAD; // Set gSerial.EE_ADRESS[0] to EEPROM is written
        gSerial.EE_ADRESS[13] = 0xF0;      // Set UI Nr. to 15;

        // 4 CHAR numbers to 1 HEX number
        // (MSB)LFT4 to (LSB)LFT1
        // LFT_1 = ((LFT1-48)+10*(LFT2-48)+100*(LFT3-48)+1000*(LFT4-48)) & 255
        LFT1_h =     ((uint16_t)(gSerial.EE_BYTE.LFT1-48)+10*(uint16_t)(gSerial.EE_BYTE.LFT2-48)
                +100*(uint16_t)(gSerial.EE_BYTE.LFT3-48)+1000*(uint16_t)(gSerial.EE_BYTE.LFT4-48))& 0xFF;

        // LFT_2 = ((LFT1-48)+10*(LFT2-48)+100*(LFT3-48)+1000*(LFT4-48)) / 256
        LFT2_h =     ((uint16_t)(gSerial.EE_BYTE.LFT1-48)+10*(uint16_t)(gSerial.EE_BYTE.LFT2-48)
                +100*(uint16_t)(gSerial.EE_BYTE.LFT3-48)+1000*(uint16_t)(gSerial.EE_BYTE.LFT4-48)) >> 8;

        BAU_h = ( (gSerial.EE_BYTE.BAU1-48)+10*(gSerial.EE_BYTE.BAU2-48) );
        FAM_h = ( (gSerial.EE_BYTE.FAM1-48)+10*(gSerial.EE_BYTE.FAM2-48) );

        gSerial.EE_BYTE.LFT_1 = (uint8_t)LFT1_h;
        gSerial.EE_BYTE.LFT_2 = (uint8_t)LFT2_h;
        gSerial.EE_BYTE.BAU_1 = BAU_h;
        gSerial.EE_BYTE.FAM_1 = FAM_h;

        array_len = sizeof(gSerial.EE_ADRESS);
        for(i = 0; i < array_len; i++)
        {

            EEADR = i;                        //Address to write
            EEDATA = gSerial.EE_ADRESS[i];    //Data to write
            EECON1bits.EEPGD = 0;             //Selecting EEPROM Data Memory
            EECON1bits.CFGS = 0;              // access eeprom //
            EECON1bits.WREN = 1;              //Enable writing of EEPROM
            INTCON_SAVE=INTCON;               //Backup INCON interupt register
            INTCON=0;                         //Diables the interrupt
            EECON2=0x55;                      //Required sequence for write to internal EEPROM
            EECON2=0xAA;                      //Required sequence for write to internal EEPROM
            EECON1bits.WR = 1;                //Initialise write cycle
            INTCON = INTCON_SAVE;             //Enables Interrupt
            EECON1bits.WREN = 0;              //To disable write


            while(PIR4bits.EEIF == 0)     //Checking for complition of write operation
            {

            }
            PIR4bits.EEIF = 0;            //Clearing EEIF bit
        }

  }

}
