/*********************************************************************
 *
 *                  BMS_CAN Header
 *
 *********************************************************************
 * FileName:        BMS_CAN.h
 * Processor:       PIC18F45K80
 * Complier:        Microchip C18 v3.41
 * Company:         KIT - CN - IPE
 *
 * Author           Date         Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Ott W.           08.02.2016   Release
 *********************************************************************/

#ifndef BMS_UI_CAN_H          // To avoid duplicate inclusion
#define BMS_UI_CAN_H



/*********************************************************************
 *
 * General purpose typedef's
 *
 ********************************************************************/
typedef union _CAN_CONFIG
{
  struct
  {
    uint8_t   SlaveNo;
    uint8_t   BRP;
    uint8_t   PROPSEG;
    uint8_t   PHSEG1;
    uint8_t   PHSEG2;
    uint8_t   SJW;
    uint8_t   PHSEG2_MODE;
    uint8_t   BUS_SAMPLE_MODE;
    uint8_t   WAKEUP_MODE;
    uint8_t   FILTER_MODE;
  };
} CAN_CONFIG;


typedef union _CAN_MAILBOX
{
  struct
  {
    uint8_t   BOX_No;
    uint32_t  CAN_Id;
    uint8_t   DLC;
    uint32_t  ACCEPTANCE_FILTER;
    int8_t    dataC[8];
  };
} CAN_MAILBOX;



/*********************************************************************
*
*                             Defines
*
*********************************************************************/
#define CAN_OK              1
#define CAN_NOK             0
#define CAN_Tx_TimeOut     50
#define LTC_TimeOut_Max     2



/*********************************************************************
 *
 * external Globals
 *
 ********************************************************************/
extern uint16_t     gVoltage;
extern int16_t      gCurrent;
extern int16_t      gOffset;
extern uint8_t      gCounter;
extern uint8_t      UIMode;
extern uint8_t      UI_ERROR;
extern uint8_t      Write_Flag;
extern uint8_t      Check_Can_Flag;
extern CAN_CONFIG   gCAN_CONFIG;

extern uint8_t      Alive_Slave[160];
extern uint8_t      Check_Alive_Slave[160];
extern uint8_t      Count_Error_Slave[160];
extern uint8_t      Alive_Master[16];
extern uint8_t      Check_Alive_Master[16];
extern uint8_t      Count_Error_Master[16];
extern uint8_t      gERROR_Counter;



/*********************************************************************
 *
 * Funktions Prototypen
 *
 ********************************************************************/
int8_t CAN_Init              ( CAN_CONFIG  *config );
int8_t CAN_Init_Mailbox      ( CAN_MAILBOX *mbox);
int8_t CAN_Write             ( CAN_MAILBOX *mbox, uint8_t *dataC );
int8_t CAN_Read              ( CAN_MAILBOX *mbox, uint8_t *data );

int8_t CAN_Write_UI          ( void );
int8_t CAN_Read_Master       ( void );
int8_t CAN_Read_Master_Slave ( void );

void   CAN_Slave_check       ( uint32_t can_id, uint8_t *ptr);
void   CAN_Master_check      ( uint32_t can_id, uint8_t *ptr);

void   CAN_Alive             ( void );

int8_t Load_Mailbox          ( uint8_t SlaveNo );
int8_t CAN_Write_System      ( void );
int8_t CAN_Read_System       ( void );



#endif
